from logicqubit import logic
from logicqubit import gates
from logicqubit import hilbert
from logicqubit import utils
from logicqubit import circuit
from logicqubit import zhegalkin
from logicqubit import oracle
from logicqubit import synthesis

__version__ = "1.5.2"

print("logicqubit-gpu version "+__version__)
